#ifndef __EXAMPLE_ENUMS_H__
#define __EXAMPLE_ENUMS_H__

G_BEGIN_DECLS

typedef enum {
	EXAMPLE_VERBOSITY_ERRORS,
	EXAMPLE_VERBOSITY_MINIMAL,
	EXAMPLE_VERBOSITY_DETAILED,
	EXAMPLE_VERBOSITY_DEBUG,
} ExampleVerbosity;

G_END_DECLS

#endif /* __EXAMPLE_ENUMS_H__ */
