#include <stdio.h>

int main (int argc, char * argv[])
{
  const char *out = "SUCCESS!";

  if (argc != 2) {
    printf ("%s\n", out);
  } else {
    int ret;
    FILE *f = fopen (argv[1], "w");
    ret = fwrite (out, sizeof (out), 1, f);
    if (ret != 1)
      return -1;
  }
  return 0;
}
