#include<stdio.h>
#include<stdlib.h>

#include<impl.h>

int main(int argc, char **argv) {
    char *ten = malloc(10);
    if(getenv("TEST_ENV")) {
        do_nasty(ten);
        printf("TEST_ENV is set.\n");
    }
    free(ten);
    return 0;
}
