#!/usr/bin/env python3
import os
from gi.repository import Gio

if __name__ == '__main__':
    res = Gio.resource_load(os.path.join('resources', 'simple-resources.gresource'))
    Gio.Resource._register(res)

    data = Gio.resources_lookup_data('/com/example/myprog/res1.txt', Gio.ResourceLookupFlags.NONE)
    assert(data.get_data() == b'This is a resource.\n')
